/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Copyright (C) 2004,2019,2024 Erik Auerswald
 * auerswal <at> unix-ag <dot> uni-kl <dot> de */

#include <errno.h>          /* used by perror */
#include <fcntl.h>          /* open */
#include <getopt.h>         /* getopt */
#include <glob.h>           /* glob */
#include <stdio.h>          /* puts, printf, BUFSIZ, perror */
#include <stdlib.h>         /* exit */
#include <string.h>         /* strcpy */
#include <unistd.h>         /* close */
#include <linux/joystick.h> /* joystick driver specific definitions */
#include <sys/ioctl.h>      /* ioctl */
#include <sys/types.h>      /* open */
#include <sys/stat.h>       /* open */
#include "lsjoystick.h"     /* defines etc. */

/* print usage */
void usage(char *name, FILE *f)
{
  fprintf(f,"Usage: %s [OPTION]... [DEVICE]...\n", name);
  fprintf(f,"\nOptions:\n");
  fprintf(f,"  -h, --help              print this help\n");
  fprintf(f,"  -V, --version           print version information\n");
  fprintf(f,"  -l, --long              print additional info, not just device and name\n");
  fprintf(f,"  -v, --verbose           explain what is beeing done\n");
  fprintf(f,"  -q, --quiet             suppress error messages\n");
  fprintf(f,"  -b, --device-base=BASE  use BASE instead of "
            DEFAULT_DEVICE_BASE "\n");
  fprintf(f,"  -c, --count             count joysticks (no info about individual joysticks)\n");
  fprintf(f,"  -g, --use-glob          query " DEFAULT_DEVICE_BASE "*\n");
  fprintf(f,"\nIf no devices are specified %s tries to find joysticks,", name);
  fprintf(f,"\notherwise the given DEVICEs are probed.\n");
}

/* get and print information about device */
void query_device(int fd, char *device, int operation)
{
  int js_driver_version;
  char error_message[BUFSIZ], js_name[BUFSIZ];
  char js_num_axes, js_num_buttons;
  /* query joystick name */
  if(operation & VERBOSE)
    printf("getting joystick name from %s\n", device);
  if(ioctl(fd, JSIOCGNAME(BUFSIZ), js_name) < 0)
  {
    if(!(operation & NO_ERRORS))
    {
      snprintf(error_message, BUFSIZ, "Error getting joystick name from %s",device);
      perror(error_message);
    }
    strncpy(js_name, "Unknown", BUFSIZ-1); js_name[BUFSIZ-1]='\0';
  }
  /* query number of axes */
  if(operation & VERBOSE)
    printf("getting number of axes from %s\n", device);
  if(ioctl(fd, JSIOCGAXES, &js_num_axes) < 0)
  {
    if(!(operation & NO_ERRORS))
    {
      snprintf(error_message, BUFSIZ, "Error getting number of joystick axes from %s",device);
      perror(error_message);
    }
    js_num_axes = 0;
  }
  /* query number of buttons */
  if(operation & VERBOSE)
    printf("getting number of buttons from %s\n", device);
  if(ioctl(fd, JSIOCGBUTTONS, &js_num_buttons) < 0)
  {
    if(!(operation & NO_ERRORS))
    {
      snprintf(error_message, BUFSIZ, "Error getting number of joystick buttons from %s",device);
      perror(error_message);
    }
    js_num_buttons = 0;
  }
  /* query driver version */
  if(operation & VERBOSE)
    printf("getting driver version from %s\n", device);
  if(ioctl(fd, JSIOCGVERSION, &js_driver_version) < 0)
  {
    if(!(operation & NO_ERRORS))
    {
      snprintf(error_message, BUFSIZ, "Error getting version number of joystick driver from %s",device);
      perror(error_message);
      fprintf(stderr,"Joystick driver is not version 1.0+\n");
    }
  }

  /* print joystick information */
  if(operation & VERBOSE)
    puts("printing joystick info if not supressed");
  if(operation & SHORT_FORMAT)
  {
    printf("%s: %s\n", device, js_name);
    if(operation & LONG_FORMAT)
      printf("%*s(%d axes, %d buttons, driver version %d.%d.%d)\n",(int) strlen(device)+4, "", js_num_axes, js_num_buttons, js_driver_version >> 16, (js_driver_version >> 8) & 0xff, js_driver_version & 0xff);
  }
}

int main(int argc, char *argv[])
{
  int i, c, operation=SHORT_FORMAT, fd, verbose=0, quiet=0,
      number_of_joysticks=0, ret;
  char device_base[BUFSIZ], globpat[BUFSIZ], device[BUFSIZ],
       error_message[BUFSIZ];
  size_t s;
  glob_t globbuf;

  device_base[BUFSIZ-1] = '\0';
  globpat[BUFSIZ-1] = '\0';
  device[BUFSIZ-1] = '\0';
  error_message[BUFSIZ-1] = '\0';

  strncpy(device_base, DEFAULT_DEVICE_BASE, BUFSIZ-1);

  /* parse command line */

  /* the following is (basically) taken from the getopt(3) manpage
   * and modified for my needs */

  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"version", 0, 0, 'V'},
      {"long", 0, 0, 'l'},
      {"quiet", 0, 0, 'q'},
      {"verbose", 0, 0, 'v'},
      {"device-base", 1, 0, 'b'},
      {"count", 0, 0, 'c'},
      {"use-glob", 0, 0, 'g'},
      {0, 0, 0, 0}
    };

    c = getopt_long (argc, argv, "hVlqvb:cg", long_options, &option_index);
    if (c == -1)
      break;

    switch (c) {
      case 'h':
	usage(argv[0],stdout);
	exit (EXIT_SUCCESS);
	break;

      case 'V':
	printf("%s version %s\n", argv[0], VERSION);
	exit (EXIT_SUCCESS);
	break;

      case 'l':
	operation &= ~QUIET_FORMAT;
	operation |= (SHORT_FORMAT | LONG_FORMAT);
	break;

      case 'q':
	quiet = 1;
	break;

      case 'v':
	verbose=1;
	operation |= VERBOSE;
	break;

      case 'b':
	if(optarg)
	  strncpy(device_base, optarg, BUFSIZ-1);
	break;

      case 'c':
	operation &= ~(SHORT_FORMAT | LONG_FORMAT);
	operation |= QUIET_FORMAT;
	break;

      case 'g':
        operation |= USE_GLOB;
	break;

      case '?':  /* missing argument or character not in optstring */
	usage(argv[0],stderr);
	exit (EXIT_FAILURE);
	break;

      default:   /* this should not be reached */
	printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }

  /* end of code from man page */

  /* iterate over joystick devices */
  if(operation & USE_GLOB) {
    strncpy(globpat, device_base, BUFSIZ-1);
    strncat(globpat, "*", BUFSIZ-1);
    if(verbose == 1) {
      printf("iterating over %s\n", globpat);
    }
    ret = glob(globpat, 0, NULL, &globbuf);
    switch(ret) {
      case GLOB_NOSPACE:
        if(!quiet) {
	  puts("glob(): out of memory");
	}
	exit(EXIT_FAILURE);
	break;

      case GLOB_ABORTED:
        if(!quiet) {
	  puts("glob(): read error");
	}
	exit(EXIT_FAILURE);
	break;

      case GLOB_NOMATCH: /* no joystick devices found */
        break;

      case 0: /* no error */
        break;

      default:
        if(!quiet) {
	  puts("glob(): unknown error");
	}
	exit(EXIT_FAILURE);
    }
    /* now globbuf contains at least one joystick file name */
    for(s=0; s<globbuf.gl_pathc; s++) {
      /* open joystick device */
      if(verbose == 1)
	printf("trying to open joystick device %s\n", globbuf.gl_pathv[s]);
      fd = open(globbuf.gl_pathv[s], O_RDONLY);
      if(fd < 0) {
	if((verbose == 1 && quiet == 0)) {
	  snprintf(error_message, BUFSIZ, "Error opening %s",
	           globbuf.gl_pathv[s]);
	  perror(error_message);
	}
      } else {
	if(verbose == 1) {
	  printf("found joystick device %s\n", globbuf.gl_pathv[s]);
	}
	number_of_joysticks++; /* found another joystick */
      /* print information about joystick connected to current device */
	if(verbose == 1 && fd > 0) {
	  printf("querying joystick device %s\n", globbuf.gl_pathv[s]);
	}
	query_device(fd, globbuf.gl_pathv[s], operation);
      /* close joystick device */
	if(verbose == 1 && fd > 0) {
	  printf("closing joystick device %s\n", globbuf.gl_pathv[s]);
	}
	if(close(fd) < 0) {
	  if(quiet == 0) { /* error messages can be suppressed */
	    snprintf(error_message, BUFSIZ, "Error closing %s",
	             globbuf.gl_pathv[s]);
	    perror(error_message);
	  }
	}
      }
    }
  } else {
    if(verbose == 1)
      puts("iterating over joystick devices");
    for(i=0; 1; i++) {
      if(optind < argc) { /* devices specified on command line */
	if(verbose == 1)
	  puts("using device from commandline");
	if(optind+i+1 <= argc) { /* not all devices tried yet */
	  strncpy(device, argv[optind+i], BUFSIZ-1);
	}
	else /* tried every device from command line */
	  break;
      } else {
	if(verbose == 1)
	  puts("constructing devicename from device_base");
	snprintf(device, BUFSIZ, "%s%d", device_base, i);
      }
      /* open joystick device */
      if(verbose == 1)
	printf("trying to open joystick device %s\n", device);
      fd = open(device, O_RDONLY);
      if(fd < 0) {
	/* this error condition is used to determin the number of joysticks,
	 * so normally no error message should be printed
	 * if devices are specified via command line this is a reportworthy
	 * error */
	if((verbose == 1 && quiet == 0) || (optind < argc)) {
	  snprintf(error_message, BUFSIZ, "Error opening %s", device);
	  perror(error_message);
	}
	if(optind >= argc)
	  break; /* exit iteration if device can't be opened */
      }
      if(verbose == 1 && fd > 0)
	printf("found joystick device %s\n", device);
      if(fd > 0)
	number_of_joysticks++; /* found another joystick */
      /* print information about joystick connected to current device */
      if(verbose == 1 && fd > 0)
	printf("querying joystick device %s\n", device);
      if(fd > 0)
	query_device(fd, device, operation);
      /* close joystick device */
      if(verbose == 1 && fd > 0)
	printf("closing joystick device %s\n", device);
      if(fd > 0) {
	if(close(fd) < 0) {
	  if(quiet == 0) { /* error messages can be suppressed */
	    snprintf(error_message, BUFSIZ, "Error closing %s",device);
	    perror(error_message);
	  }
	}
      }
    }
  }
  /* print number of found joysticks */
  if(operation & (QUIET_FORMAT | LONG_FORMAT)) {
    if(number_of_joysticks != 1) {
      printf("%d joysticks\n", number_of_joysticks);
    } else {
      puts("1 joystick");
    }
  }
  exit (EXIT_SUCCESS);
}
